# -*- coding: utf-8 -*-
"""
sample_GetMetData-Hourly_a.py

時別値取得のサンプルプログラム:
  １メッシュの時系列データを抽出
  地点は、lalodomainで指定する。
  20日前から、15日先までの範囲のグラフが作図される。（要、matplotlib環境）
  日別値は、日平均、日最高、日最低が、
　時別値は、予報期間の9日先までがプロットされる。



### 時別値メッシュについて -----------------------------------------------------
時刻指定は、"YYYY-MM-DDThh"　のように、YYYY-MM-DD　の後ろに
Tを挟んで時間の数値を二桁で加える。

時間指定をしない場合は、対象の日付の丸々1日分を指定したと解釈される。

例１　　timedomain=「'2020-10-10','2020-10-10'」  
        →　2020年10月10日 01:00 ~ 24:00　を取得する。
                
時別値について補足：
　過去値は、毎正時のアメダスデータから作成される。日別値の最高・最低気温は10分値ベースなので、
 時別値の範囲とは若干異なる
 
 　予報値は、GPVから得られる時別値全体に、メッシュ日別値を作成する際の
   日平均気温と同じ補正量を施している。そのため、日最高・最低気温にはズレが生じるが、
　　メッシュ日別値の平均気温と、メッシュ時別値24個から計算する日平均気温とは一致する。
--------------------------------------------------------------------------

NARO/HARC NEMOTO Manabu, 2021.01.08


"""
import AMD_Tools3 as AMD
import datetime
import matplotlib.pyplot as plt
import matplotlib.dates as mdates


#　位置指定
lalodomain = [36.00417, 36.00417, 140.0063, 140.0063]

# 時間範囲指定
TODAY = datetime.datetime.today()
SDATE = TODAY - datetime.timedelta(days=20)
EDATE = TODAY + datetime.timedelta(days=15)

timedomain = [ SDATE.strftime('%Y-%m-%d'), EDATE.strftime('%Y-%m-%d') ] 


# 気象データの取得
TMPmea_d, tim_d, lat, lon = AMD.GetMetData('TMP_mea', timedomain, lalodomain)
TMPmax_d, tim_d, lat, lon = AMD.GetMetData('TMP_max', timedomain, lalodomain)
TMPmin_d, tim_d, lat, lon = AMD.GetMetData('TMP_min', timedomain, lalodomain)
TMP_h, tim_h, lat, lon = AMD.GetMetDataHourly('TMP', timedomain, lalodomain)

TMPmea_d = TMPmea_d[:,0,0]
TMPmax_d = TMPmax_d[:,0,0]
TMPmin_d = TMPmin_d[:,0,0]
TMP_h    = TMP_h[:,0,0]

print(TMP_h)


## 作図

x_h = tim_h
y_h = TMP_h

x_d = tim_d
y_d1 = TMPmea_d
y_d2 = TMPmax_d
y_d3 = TMPmin_d

fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
ax.plot(x_h,y_h, label='hourly')
ax.plot(x_d,y_d1, label='daily mean')
ax.plot(x_d,y_d2, label='daily max')
ax.plot(x_d,y_d3, label='daily min')

##以下をカスタマイズする
# daysFmt = mdates.DateFormatter('%Y-%m-%d')
daysFmt = mdates.DateFormatter('%m/%d')
ax.xaxis.set_major_formatter(daysFmt)
fig.autofmt_xdate()




# plt.plot(x_h,y_h, label='hourly')
# plt.plot(x_d,y_d1, label='daily mean')
# plt.plot(x_d,y_d2, label='daily max')
# plt.plot(x_d,y_d3, label='daily min')


#plt.xticks(np.arange(min(x_d), max(x_d), 5))
plt.legend(loc='best')
plt.savefig('result_a'+'.png', dpi=400)
plt.show()