# -*- coding: utf-8 -*-
"""
sample_GetMetData-Hourly_b.py

ある時刻における、空間分布描画のサンプルプログラム:
  
  範囲は、lalodomainで指定する。
  時刻は、timedomainで指定する。
  指定範囲＆時刻の気温空間分布が作図される（要、matplotlib環境）。
　カラースケールは、　manualsclをFalse指定すると、自動で割り当てられる。
  manualscl=Trueの場合は、Sclmax, Sclmin, Sclint で調整する。
  

### 時別値メッシュについて -----------------------------------------------------
時刻指定は、"YYYY-MM-DDThh"　のように、YYYY-MM-DD　の後ろに
Tを挟んで時間の数値を二桁で加える。

時間指定をしない場合は、対象の日付の丸々1日分を指定したと解釈される。

例１　　timedomain=「'2020-10-10','2020-10-10'」  
        →　2020年10月10日 01:00 ~ 24:00　を取得する。
                
時別値について補足：
　過去値は、毎正時のアメダスデータから作成される。日別値の最高・最低気温は10分値ベースなので、
 時別値の範囲とは若干異なる
 
 　予報値は、GPVから得られる時別値全体に、メッシュ日別値を作成する際の
   日平均気温と同じ補正量を施している。そのため、日最高・最低気温にはズレが生じるが、
　　メッシュ日別値の平均気温と、メッシュ時別値24個から計算する日平均気温とは一致する。
--------------------------------------------------------------------------

NARO/HARC NEMOTO Manabu, 2021.01.08


"""
import AMD_Tools3 as AMD
import matplotlib.pyplot as plt
import numpy as np

#　位置指定
lalodomain = [ 42.75, 44.75, 141.0, 143.0 ]

# 時間範囲指定     日付の後ろに「T」を挟んで、時刻を二桁で加える。                         
timedomain = [ '2020-05-20T06', '2020-05-20T06'] 
 
# 気象データの取得
TMP_h, tim_h, lat, lon, nam, uni = AMD.GetMetDataHourly('TMP', timedomain, lalodomain, namuni=True )
TMP_h    = TMP_h[0,:,:]

print(TMP_h)


# 作図　
D2D = TMP_h

tate = 6
figtitle = nam+" ["+uni+"] on "+tim_h[0].strftime('%Y-%m-%d %Hh')
manualscl = False

sclmax = 16
sclmin = -6
sclint = 2.0
if not manualscl :
    sclmax = round(np.nanmax(np.array(D2D)),1)
    sclmin = round(np.nanmin(np.array(D2D)),1)
    sclint = round(((sclmax-sclmin)/10.0),1)

levels = np.arange(sclmin, sclmax+sclint, sclint)
yoko = tate * (np.max(lon)-np.min(lon))/(np.max(lat)-np.min(lat)) + 2 
fig = plt.figure(num=None, figsize=(yoko, tate))

plt.axes(facecolor='0.8')
levels = np.arange(sclmin, sclmax+sclint, sclint)
cmap = plt.cm.Spectral_r
cmap.set_over('w', 1.0)
cmap.set_under('k', 1.0)
CF = plt.contourf(lon, lat, D2D, levels, cmap=cmap, extend='both')
plt.colorbar(CF)
plt.title(figtitle)
plt.savefig('result_b'+'.png', dpi=400)
plt.show()
plt.clf()

