# -*- coding: utf-8 -*-
"""
 滋賀県を対象に2017年6月1日の水田水温（日平均水温）を計算して地図上に描画するサンプルプロラム
です。プログラムを実行すると図のファイルPadTemp.htmlが作成されます。

 水田水温は、植生がなく水深50mmの場合の水温を計算しています。発育ステージや水深を変えるにはdvs
やDwを変更してください。

2018.02.16 Atsushi Maruyama
"""
import AMD_Tools3 as AMD
from AMD_PaddyWaterTemp import WaterTemp
import numpy as np

timedomain = ["2017-06-01", "2017-06-01"] #期間の設定。
lalodomain  = [34.783, 35.708, 135.765, 136.460] #範囲の設定。
Prf, lat, lon = AMD.GetGeoData("pref_2500", lalodomain) #都道府県（滋賀県）の範囲
Pad, lat, lon = AMD.GetGeoData("landuse_H210100", lalodomain) #土地利用（水田割合）
PadAri = np.where(Pad<0.01, np.nan, 1.0)

dvs = [0]  #期間中の発育ステージ(期間分のリストであること)
Dw  = [50] #期間中の水深(期間分のリストであること)
Tmea, Tmax, Tmin, tim, lat, lon = WaterTemp(timedomain, lalodomain, dvs, Dw)
Tmea = Tmea[0, :, :]
Map = Tmea * PadAri * Prf  #対象の都道府県、土地利用のみを描画

AMD.PutGSI_Map(Map,lat,lon,label="Paddy Water Temperature [degC]",
               filename="PadTemp")
